/*
  Macro           HistCleaner       (formerly named "DelNonExistEditHist")
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   TSE v3 upwards, all variants.
  Version         v2   30 Oct 2023


  This extension deletes non-existing files from TSE's edit history.

  Its main purpose is to do so once per TSE session at TSE's start-up,
  but you can also load or execute it at will.

  It runs invisibly in the background, possibly up to a few minutes.
  You can keep editing while it runs.
  It does not block TSE, with one possible and documented exception.
  It stops after the whole edit history has been checked.

  Default it will always pop up a list of deletable entries when ready.
  By executing the extension as a macro you can configure it to show this list
  always, if there are deletions, or never.
  In the last case it will perform the deletions silently.

  Default it will delete fewer entries than it can: See the CAVEAT section.
  You can configure it to delete more by executing the extension.

  TSE's edit history is used by these TSE functions:
  EditFile(), EditThisFile(), InsertFile(), SaveBlock(), and SaveAs().
  See TSE's built-in documentation for more info on them.


  CAVEAT

  The extension's functionality has 3 logical problems.
  Two are avoided by default turning a part of the functionality off.
  The third one is mitigated, not solved.

  1.
    An unqualified file is one with out a drive (Windows) or root (Linux)
    or without a share (starting with "\\" in Windows).

    An unqualified file exists or does not exist based on TSE's current
    directory.

    Depending on how a user starts TSE and on TSE's PickFileChangesDir setting,
    TSE's current directory can change, thereby making unqualified files
    sometimes exist and sometimes not exist.

    Only you can decide on how to handle this in your situation.

    Therefore, based on a configuration option that is default OFF, this
    extension will or will not delete the edit history entries of unqualified
    files that do not exist relative to TSE's current directory.

  2.
    Implemented for Windows only:

    Some of us deal with network drives or shares that do not always exist.
    A typical use case is a laptop that is used at work and at home:
    When connected to the work environment, network drives exist that do not
    exist in the home environment.
    There are many other cases in which a network drive or share may only
    occasionally "exist" or "be connected".

    When a network drive or share is disconnected, then all files on it do not
    exist as far as TSE is concerned.
    So its edit history entries would be deletable, but that is not always what
    we want.

    Only you can decide on how to handle this in your situation.

    Therefore, based on a configuration option that is default OFF, this
    extension will or will not delete the edit history entries of fully
    qualified files on drives or shares that are currently disconnected
    or otherwise do not exist.

  3.
    Accessing a network or share is typically slow for the first access,
    relatively fast for each subsequent access, and reverts to another
    slow first access when the network has not been accessed for a while.

    When this extension checks TSE's edit history and encounters a network
    or share for the first time in the TSE session, then the above
    causes TSE to briefly freeze.

    Or not so briefly if a network has an exceptionally slow initial connection.
    ( Interestingly, accessing a local WSL instance from Windows behaves like a
      network connection too. For me this causes an acceptably short freeze. )

    This extension mitigates this annoyance in two ways:
    - It groups and sorts directories and files by Windows drive and share.
      This ensures that a freeze occurs at most once per network/share.
    - Networks and shares are checked last.
      This ensures that no freeze occurs immediately after TSE's start-up.
    That said, a mitigation is not a fix, so if you have edited files on a
    network or share with a horribly slow initial connection, then it does not
    make sense to use this extension as an autoloaded macro.
    It then does still make sense to occasionally load or execute the extension
    yourself to clean up your edit history.


  INSTALLATION

  Copy this file to TSE's "mac" folder and compile it there, for instance by
  opening the file there in TSE and applying the Macro -> Compile menu.

  Optional, advised:
    Add the macro's name "HistCleaner" to TSE's Macro AutoLoad list.
    This runs the extension silently each time TSE is started up.
    Alternatively you can load or execute the macro at will to run it.

  Optional:
    Execute "HistCleaner" as a macro to configure it.

  Possibly:
    Version 1 (20 Oct 2023) of this extension was called "DelNonExistHist".
    If you added that name to your Macro AutoLoad List then you need to delete
    that name from there.


  TODO
    MUST
    SHOULD
    COULD
    WONT
      The possibly slow response of a network or share is a problem that is
      neither limited nor specific to this extension. I am contemplating an
      independent solution that keeps network and share connections responsive
      for all tools and users.


  HISTORY

  v1      20 Oct 2023
    Initial release under the name "DelNonExistHist".

  v1.0.0.1    25 Oct 2023
  - Renamed the extension to "HistCleaner".
  - Mitigated (not solved) TSE briefly freezing when a network file is checked:
    - Entries containing a network reference are grouped sequentially,
      causing at most one freeze per network.
    - Entries containing a network reference are checked last,
      so no freeze happens immediately after TSE's start-up.
  - "*" is now allowed in an entry and is checked for existence.
  - An entry that enumerates multiple directories and/or files is deleted from
    history if any of them does not exist.
  - Use of "-e" and a macro name deletes its edit history entry if the macro
    does not exist.
  - An existing single unquoted directory or file with spaces in its name
    is no longer deleted.
  - The result list can now be configured to be shown always, if there are
    deletions, or never.
  - If the result list is configured to be shown, then it lets you accept or
    reject the deletions, otherwise the deletions are done at detection time.

  v1.0.0.2    27 Oct 2023
  - Added viewing the documentation from the configuration menu.
  - Added configuration menu options for unqualified files and disconnected
    drives and shares, but these configuration options do not do anything yet.

  v1.0.0.3    27 Oct 2023
  - The configuration options now work.
  - Executing the macro to configure it will no longer implcitly run it.

  v1.0.0.4    28 Oct 2023
  - Added Linux compatibility.

  v2          30 Oct 2023
  - Broad release.
  - Updated documentation.
  - Removed test code.

*/





// Start of compatibility restrictions and mitigations

#ifndef EDITOR_VERSION
   Editor Version is older than TSE 3.0. You need at least TSE 3.0.
#endif

// End of compatibility restrictions and mitigations.





// Global constants and semi-constants
#define CCF_OPTIONS                       _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_
string  CFG_SECTION_NAME [MAXSTRINGLEN] = ''
string  DELETABLES_LIST_FOOTER     [65] =  '<Enter> Delete all these entries, <Alt E> Edit list, or <Escape>.'
integer DEL_ID                          = 0
integer HST_ID                          = 0
string  MACRO_NAME       [MAXSTRINGLEN] = ''
integer MACRO_LOAD_TIME                 = 0
#define REPORT_TYPE_LENGTH                23
integer SRT_ID                          = 0


// Global variables

#ifndef LINUX
  integer cfg_del_disconnecteds = FALSE
#endif

integer cfg_del_unqualifieds  = FALSE
integer cfg_report_type       = 0
integer edit_deletables_list  = FALSE
integer edit_id               = 0
integer entries_checked       = 0
integer entry_to_check        = 1
integer indexes_deleted       = 0
integer is_macro_running      = TRUE
integer run_macro             = FALSE


integer proc write_profile_int(string  section,
                               string  item_name,
                               integer item_value)
  integer result = WriteProfileInt(section,
                                   item_name,
                                   item_value)
  if not result
    Alarm()
    Warn(MACRO_NAME; 'error: Failed to write configuration to file "tse.ini".')
  endif
  return(result)
end write_profile_int


integer proc silent_CopyBlock()
  integer old_MsgLevel = Set(MsgLevel, _WARNINGS_ONLY_)
  integer result       = FALSE
  result = CopyBlock()
  Set(MsgLevel, old_MsgLevel)
  return(result)
end silent_CopyBlock


#ifndef LInux
  integer proc silent_Sort(integer flag)
    integer old_MsgLevel = Set(MsgLevel, _WARNINGS_ONLY_)
    integer result       = FALSE
    result = Sort(flag)
    Set(MsgLevel, old_MsgLevel)
    return(result)
  end silent_Sort
#endif

/*
  Mote:
    hlp_id is intentionally permanent once created.
    If a user requests the help and then purges the macro,
    then that does not mean that the users's requested view on the
    documentation should be closed.
*/
proc show_help()
  string  full_macro_source_name [MAXSTRINGLEN] = SplitPath(CurrMacroFilename(), _DRIVE_|_PATH_|_NAME_) + '.s'
  string  help_file_name         [MAXSTRINGLEN] = '*** ' + MACRO_NAME + ' Help ***'
  integer hlp_id                                = GetBufferId(help_file_name)
  integer org_id                                = GetBufferId()
  integer tmp_id                                = 0
  if hlp_id
    GotoBufferId(hlp_id)
    UpdateDisplay()
  else
    tmp_id = CreateTempBuffer()
    if LoadBuffer(full_macro_source_name)
      // Separate characters, otherwise my SynCase macro gets confused.
      if lFind('/' + '*', 'g')
        PushBlock()
        UnMarkBlock()
        Right(2)
        MarkChar()
        if not lFind('*' + '/', '')
          EndFile()
        endif
        MarkChar()
        Copy()
        CreateTempBuffer()
        Paste()
        UnMarkBlock()
        PopBlock()
        BegFile()
        ChangeCurrFilename(help_file_name, CCF_OPTIONS)
        BufferType(_NORMAL_)
        FileChanged(FALSE)
        BrowseMode(TRUE)
        UpdateDisplay()
      else
        GotoBufferId(org_id)
        Warn('File "', full_macro_source_name, '" has no multi-line comment block.')
      endif
    else
      GotoBufferId(org_id)
      Warn('File "', full_macro_source_name, '" not found.')
    endif
    AbandonFile(tmp_id)
  endif
end show_help


integer proc has_s_option(string s)
  integer token_no = 0

  for token_no = 1 to NumFileTokens(s)
    if EquiStr(GetFileToken(s, token_no), '-s')
      return(TRUE)
    endif
  endfor
  return(FALSE)
end has_s_option


string proc strip_single_file_options(string s)
  string result [MAXSTRINGLEN] = s

  // "-b -x my file."

  while Length(result)
    case result[1]
      when ' '
        result = result[2: MAXSTRINGLEN]
      when '-'
        if (Lower(result[2]) in 'k', 'l', 'e', 'n')
        and       result[3]  == ' '
          result = Trim(result[4: MAXSTRINGLEN])
        endif
        while not (result[1] in ' ', '')
          result = result[2: MAXSTRINGLEN]
        endwhile
      otherwise
        break
    endcase
  endwhile

  return(result)
end strip_single_file_options


integer proc is_accepted_as_unqualified_or_disconnected(string token)
  integer result = FALSE

  // Note:  This proc is only called if the token does not exist
  //        as a directory or file.

  #ifdef LINUX
    if token[1] <> '/'
      // Token does not refer to a qualified drive or share.
      // Do we accept unqualified files in our TSE history?
      result = not cfg_del_unqualifieds
    endif
  #else
    if (    isAlpha(token[1])
        and token[2] == ':'  )
      // Token refers to a qualified drive.
      // Is this drive connected?
      if not FileExists(token[1:2] + '\*')
        // No, the drive is not connected.
        // Do we accept files on disconnected drives in our TSE history?
        result = not cfg_del_disconnecteds
      endif
    elseif token[1:2] == '\\'
      // Token refers to a qualified drive or share.
      // Is this share connected?
      // The way I understand it, in Windows the share's root can be named
      // "\\servername\sharename\" or "\\sharename\".
      if NumTokens(token, '\') >= 4
        if  not (FileExists('\\' + GetToken(token, '\', 3)) & _DIRECTORY_)
        and not (FileExists('\\' + GetToken(token, '\', 3) + '\' + GetToken(token, '\', 4)) & _DIRECTORY_)
          // No, the share is not connected.
          // Do we accept files on disconnected shares in our TSE history?
          result = not cfg_del_disconnecteds
        endif
      endif
    else
      // Token does not refer to a qualified drive or share.
      // Do we accept unqualified files in our TSE history?
      result = not cfg_del_unqualifieds
    endif
  #endif

  return(result)
end is_accepted_as_unqualified_or_disconnected


proc idle()
  string  e_macro_name         [MAXSTRINGLEN] = ''
  integer n                                   = 0
  integer ok_tokens                           = 0
  string  possible_single_file [MAXSTRINGLEN] = ''
  string  s                    [MAXSTRINGLEN] = ''
  string  token                [MAXSTRINGLEN] = ''
  integer token_no                            = 0

  if SRT_ID
    PushLocation()
    GotoBufferId(SRT_ID)
    if (entry_to_check in 1 .. NumLines())
      GotoLine(entry_to_check)
      s = GetText(1, MAXSTRINGLEN)
      if      Length(s)
      and not has_s_option(s)
        possible_single_file = strip_single_file_options(s)
        if not FileExists(possible_single_file)
          ok_tokens = 0
          for token_no = 1 to NumFileTokens(s)
            token = GetFileToken(s, token_no)
            if EquiStr(token[1:2], '-e')
              if Length(token) > 2
                e_macro_name = Trim(token[3:MAXSTRINGLEN])
              else
                token_no     = token_no  + 1
                e_macro_name = GetFileToken(s, token_no)
                if e_macro_name <> ''
                  ok_tokens  = ok_tokens + 1
                endif
              endif
              if SplitPath(e_macro_name, _EXT_) == ''
                e_macro_name = e_macro_name + '.mac'
              endif
              if SearchPath(e_macro_name, Query(TSEPath), 'mac') <> ''
                ok_tokens = ok_tokens + 1
              endif
            elseif     token[1] == '-'
            and    not Pos(' ', token)
              ok_tokens = ok_tokens + 1
            elseif FileExists(token)
            or     is_accepted_as_unqualified_or_disconnected(token)
              ok_tokens = ok_tokens + 1
            endif
          endfor
          if ok_tokens <> NumFileTokens(s)
            n = FindHistoryStr(s, _EDIT_HISTORY_)
            if n
              AddLine(s, DEL_ID)
              if cfg_report_type == 3  // Report not shown, can delete immediately.
                DelHistoryStr(_EDIT_HISTORY_, n)
              endif
              indexes_deleted = indexes_deleted + 1
            endif
          endif
        endif
      endif
      entry_to_check  = entry_to_check  + 1
      entries_checked = entries_checked + 1
    else
      PurgeMacro(CurrMacroFilename())
    endif
    PopLocation()
  else
    PurgeMacro(CurrMacroFilename())
  endif
end idle


string proc get_cfg_report_type()
  string result [REPORT_TYPE_LENGTH] = ''

  case cfg_report_type
    when 2
      result = 'If there are deletions'
    when 3
      result = 'Never'
    otherwise
      result = 'Always'
  endcase

  return(result)
end get_cfg_report_type


proc set_cfg_report_type()
  integer result = 0

  case cfg_report_type
    when 2
      PushKey(<Tab>)
    when 3
      PushKey(<Tab>)
      PushKey(<Tab>)
  endcase

  result = MsgBoxEx(MACRO_NAME,
                    'When finished, do you want a report?',
                    '[&Always];[&If there are deletions];[&Never]')

  if (result in 1, 2, 3)
    cfg_report_type = result
    write_profile_int(CFG_SECTION_NAME, 'ReportType', cfg_report_type)
  endif
end set_cfg_report_type


#ifndef LINUX
  string proc get_cfg_del_disconnecteds()
    return(iif(cfg_del_disconnecteds, 'On', 'Off'))
  end get_cfg_del_disconnecteds


  proc toggle_cfg_del_disconnecteds()
    cfg_del_disconnecteds = not cfg_del_disconnecteds
    write_profile_int(CFG_SECTION_NAME, 'DelDisconnecteds', cfg_del_disconnecteds)
  end toggle_cfg_del_disconnecteds
#endif


string proc get_cfg_del_unqualifieds()
  return(iif(cfg_del_unqualifieds, 'On', 'Off'))
end get_cfg_del_unqualifieds


proc toggle_cfg_del_unqualifieds()
  cfg_del_unqualifieds = not cfg_del_unqualifieds
  write_profile_int(CFG_SECTION_NAME, 'DelUnqualifieds', cfg_del_unqualifieds)
end toggle_cfg_del_unqualifieds


string proc get_macro_running_state()
  return(iif(is_macro_running, 'Already running', 'Not running yet'))
end get_macro_running_state


proc set_macro_to_run()
  run_macro = TRUE
end set_macro_to_run


integer proc get_macro_running_flags()
  return(_MF_CLOSE_BEFORE_ | iif(is_macro_running, _MF_GRAYED_|_MF_SKIP_, _MF_ENABLED_))
end get_macro_running_flags


menu config_menu()
  title = 'HistCleaner configuration menu'

  '&Help ...',
    show_help(),,
    "Show the extension's documentation."

  'Rep&ort' [get_cfg_report_type():REPORT_TYPE_LENGTH],
    set_cfg_report_type(),
    _MF_DONT_CLOSE_,
    'Set when a check should report its result.'

  'Delete &unqualified files' [get_cfg_del_unqualifieds():3],
    toggle_cfg_del_unqualifieds(),
    _MF_DONT_CLOSE_,
    'Delete filenames that do not contain a drivename or sharename?'

#ifndef LINUX

  'Delete &disconnected drives and shares' [get_cfg_del_disconnecteds():3],
    toggle_cfg_del_disconnecteds(),
    _MF_DONT_CLOSE_,
    'Delete drives and shared file systems that are currently not connected?'

#endif

  '&Run ...' [get_macro_running_state():15],
    set_macro_to_run(),
    get_macro_running_flags(),
    'Run '

end config_menu


Keydef deletables_list_keys
  <Alt E>
    edit_deletables_list = TRUE
    PushKey(<Enter>)
end deletables_list_keys


proc cleanup_deletables_list()
  UnHook(cleanup_deletables_list)
  Disable(deletables_list_keys)
end cleanup_deletables_list


proc prep_deletables_list()
  UnHook(prep_deletables_list)
  ListFooter(DELETABLES_LIST_FOOTER)
  Enable(deletables_list_keys)
  Hook(_LIST_CLEANUP_, cleanup_deletables_list)
end prep_deletables_list


proc WhenPurged()
  string  edit_hist_entry [MAXSTRINGLEN] = ''
  string  msg             [MAXSTRINGLEN] = ''
  integer n                              = 0

  if (cfg_report_type in 1, 2)
    PushLocation()
    GotoBufferId(del_id)
    if cfg_report_type == 1
    or (    cfg_report_type == 2
        and NumLines())
      if Query(Beep)
        Alarm()
      endif
      Delay(5)
      while KeyPressed()
      or    KeyPressed()
        GetKey()
      endwhile
      if NumLines()
        msg = Format(MACRO_NAME, ':'; NumLines(); 'deletable edit history entries.')
        edit_deletables_list = FALSE
        Hook(_LIST_STARTUP_, prep_deletables_list)
        BegFile()
        if List(msg,
                Max(Max(LongestLineInBuffer(),
                        Length(msg) + 4),
                    Length(DELETABLES_LIST_FOOTER) + 4))
          if edit_deletables_list
            PushBlock()
            PushLocation()
            MarkLine(1, NumLines())
            edit_id = NewFile()
            silent_CopyBlock()
            PopLocation()
            PopBlock()
          else
            BegFile()
            repeat
              edit_hist_entry = GetText(1, MAXSTRINGLEN)
              n = FindHistoryStr(edit_hist_entry, _EDIT_HISTORY_)
              if n
                DelHistoryStr(_EDIT_HISTORY_, n)
              endif
            until not Down()
          endif
        endif
      else
        MsgBox(MACRO_NAME, 'There are no edit history entries to delete.', _OK_)
      endif
    endif
    PopLocation()
  endif

  if edit_id
    GotoBufferId(edit_id)
    BegFile()
  endif

  AbandonFile(DEL_ID)
  AbandonFile(HST_ID)
  AbandonFile(SRT_ID)
end WhenPurged


proc WhenLoaded()
  integer n                    = 0
  string  s     [MAXSTRINGLEN] = ''

#ifndef LINUX
  string  srt_key          [1] = ''
  string  token [MAXSTRINGLEN] = ''
  integer token_no             = 0
#endif

  MACRO_LOAD_TIME  = GetTime()
  MACRO_NAME       = SplitPath(CurrMacroFilename(), _NAME_)
  CFG_SECTION_NAME = MACRO_NAME + ':Config'

  cfg_report_type       = GetProfileInt(CFG_SECTION_NAME, 'ReportType'      , 1)
#ifndef LINUX
  cfg_del_disconnecteds = GetProfileInt(CFG_SECTION_NAME, 'DelDisconnecteds', FALSE)
#endif
  cfg_del_unqualifieds  = GetProfileInt(CFG_SECTION_NAME, 'DelUnqualifieds' , FALSE)

  PushLocation()

  DEL_ID = GetBufferId(MACRO_NAME + ':EditHistoryEntries_Deleted')
  if DEL_ID
    EmptyBuffer(DEL_ID)
  else
    DEL_ID = CreateTempBuffer()
    ChangeCurrFilename(MACRO_NAME + ':EditHistoryEntries_Deleted',
                       CCF_OPTIONS)
  endif

  SRT_ID = GetBufferId(MACRO_NAME + ':EditHistoryEntries_Check_Ordered')
  if SRT_ID
    EmptyBuffer(SRT_ID)
  else
    SRT_ID = CreateTempBuffer()
    ChangeCurrFilename(MACRO_NAME + ':EditHistoryEntries_Check_Ordered',
                       CCF_OPTIONS)
  endif

  HST_ID = GetBufferId(MACRO_NAME + ':EditHistoryEntries_At_Startup')
  if HST_ID
    GotoBufferId(HST_ID)
    EmptyBuffer()
  else
    HST_ID = CreateTempBuffer()
    ChangeCurrFilename(MACRO_NAME + ':EditHistoryEntries_At_Startup',
                       CCF_OPTIONS)
  endif

  for n = 1 to NumHistoryItems(_EDIT_HISTORY_)
    s = GetHistoryStr(_EDIT_HISTORY_, n)
    if  Length(s)
    and not has_s_option(s)
    and not Pos('://', s)
      AddLine(s)
    endif
  endfor

  if NumLines()
    PushBlock()
    MarkLine(1, NumLines())
    GotoBufferId(SRT_ID)
    silent_CopyBlock()
    UnMarkBlock()
    BegFile()
    #ifndef LINUX
      repeat
        srt_key = '!'
        s       = GetText(1, MAXSTRINGLEN)
        for token_no = 1 to NumFileTokens(s)
          token = GetFileToken(s, token_no)
          if token[1] <> '-'
            if  isAlpha(token[1])
            and token[2] == ':'
              srt_key = iif(Upper(token[1]) > srt_key, Upper(token[1]), srt_key)
            elseif token[1:2] == '\\'
              srt_key = iif('\' > srt_key, '\', srt_key)
            endif
          endif
        endfor
        BegLine()
        InsertText(srt_key, _INSERT_)
      until not Down()
      MarkLine(1, NumLines())
      silent_Sort(_IGNORE_CASE_)
      MarkColumn(1, 1, NumLines(), 1)
      KillBlock()
    #endif
    PopBlock()
  endif

  PopLocation()
  Hook(_IDLE_, idle)
end WhenLoaded


proc Main()
  is_macro_running = (GetTime() - MACRO_LOAD_TIME >= 100)
  run_macro        = FALSE

  config_menu()

  if run_macro
    Delay(9)

    case cfg_report_type
      when 2
        MsgBox(MACRO_NAME,
               Format('A check is now running in the background.', Chr(13),
                      'Only if deletions are found will a report pop up.'))
      when 3
        MsgBox(MACRO_NAME,
               Format('A check is now running in the background.', Chr(13),
                      'When ready it will exit silently.'))
      otherwise
        MsgBox(MACRO_NAME,
               Format('A check is now running in the background.', Chr(13),
                      'When ready a report will pop up.'))
    endcase
  else
    if not is_macro_running
      PurgeMacro(CurrMacroFilename())
    endif
  endif
end Main


